package dareka.common;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * eLXg߂̃[eBeBNX
 * @since NicoCache_nl+111111mod
 */
public class TextUtil {
	
	private static final Pattern VIDEO_ID_PATTERN = Pattern.compile(
			"[a-z]{2}\\d{1,9}");
	private static final Pattern THREAD_ID_PATTERN = Pattern.compile(
			"\\d{10,}");
	private static final Pattern BLANK_PATTERN = Pattern.compile(
			"[\\s@]*");
	private static final Pattern STRIP_TAGS_PATTERN = Pattern.compile(
			"<[^>]*>");
	
	/**
	 * ID(=p2{9ȉ̐)H("low"t̂͊܂܂Ȃ)
	 * 
	 * @param id 肷镶 (null)
	 * @return 񂪓ID̏ꍇ̂ true
	 **/
	public static boolean isVideoId(String id) {
		return id != null && VIDEO_ID_PATTERN.matcher(id).matches();
	}
	
	/**
	 * XbhID(=10ȏ̐̂)H
	 * 
	 * @param id 肷镶 (null)
	 * @return 񂪃XbhID̏ꍇ̂ true
	 */
	public static boolean isThreadId(String id) {
		return id != null && THREAD_ID_PATTERN.matcher(id).matches();
	}
	
	/**
	 * 󔒂݂̂̕񂩁H
	 * 
	 * @param content 肷镶 (null)
	 * @return 󔒂݂̂̕ or nullȂ true
	 */
	public static boolean isBlank(String content) {
		return content == null || BLANK_PATTERN.matcher(content).matches();
	}
	
	/**
	 * ^O(Prototype.js ̓\bhƓ)
	 * 
	 * @param content Ώە (null)
	 * @return ^O
	 */
	public static String stripTags(String content) {
		if (content == null) {
			return null;
		}
		return STRIP_TAGS_PATTERN.matcher(content).replaceAll("");
	}
	
	
	private static final Pattern ZERO_WIDTH_CHAR_PATTERN = Pattern.compile(
			"[\\u0323\\u2000-\\u200f\\u2028-\\u202e]");
	
	/**
	 * [(ZERO WITH SPACE)
	 * 
	 * @param content Ώە (null)
	 * @return [폜
	 */
	public static String stripZeroWithChars(String content) {
		if (content == null) {
			return null;
		}
		return ZERO_WIDTH_CHAR_PATTERN.matcher(content).replaceAll("");
	}
	
	/**
	 * HTML ̓ꕶGXP[v(Prototype.js ̓\bhƓ)<br>
	 * ΏۂƂȂ镶 <b>& < > " '</b> 5
	 * 
	 * @param content Ώە (null)
	 * @return GXP[v
	 * @since NicoCache_nl+120609mod
	 */
	public static String escapeHTML(String content) {
		if (content == null) {
			return null;
		}
		StringBuilder sb = new StringBuilder(content.length() * 2);
		char c;
		for (int i = 0; i < content.length(); i++) {
			switch (c = content.charAt(i)) {
			case '&':  sb.append("&amp;" ); break;
			case '<':  sb.append("&lt;"  ); break;
			case '>':  sb.append("&gt;"  ); break;
			case '"':  sb.append("&quot;"); break;
			case '\'': sb.append("&#039;"); break;
			default:   sb.append(c);
			}
		}
		return sb.toString();
	}
	
	
	private static final Pattern UNESCAPE_PATTERN = Pattern.compile(
			"&([A-Za-z]+|#(\\d+));");
	private static final HashMap<String, Integer> unescapeMap =
			new HashMap<String, Integer>();
	static {
		unescapeMap.put("amp" , 38); // &
		unescapeMap.put("lt"  , 60); // <
		unescapeMap.put("gt"  , 62); // >
		unescapeMap.put("quot", 34); // "
		unescapeMap.put("apos", 39); // '
	}
	private static String i2s(Integer i) {
		return String.valueOf((char)i.intValue());
	}
	
	/**
	 * HTML ̃GXP[vɖ߂(Prototype.js ̓\bhƓ)
	 * 
	 * @param content Ώە (null)
	 * @return ɖ߂
	 * @since NicoCache_nl+120609mod
	 */
	public static String unescapeHTML(String content) {
		if (content == null) {
			return null;
		}
		StringBuffer sb = new StringBuffer(content.length());
		Matcher m = UNESCAPE_PATTERN.matcher(content);
		String str, rep;
		while (m.find()) {
			if (unescapeMap.containsKey(str = m.group(1))) {
				rep = i2s(unescapeMap.get(str));
			} else if (str.startsWith("#")) {
				rep = i2s(Integer.valueOf(m.group(2)));
			} else {
				rep = "$0";
			}
			m.appendReplacement(sb, rep);
		}
		return m.appendTail(sb).toString();
	}
	
	/**
	 * {I XML ̎Q(amp,lt,gt,quot,apos)̕ɒu<br>
	 * {@linkplain #unescapeHTML(String)} Ƃ͈قȂA͘A
	 * <b>'&amp;amp;lt;'</b>  <b>'<'</b> ƒuB
	 * 
	 * @param content Ώە (null)
	 * @return u
	 */
	public static String replaceER(String content) {
		if (content == null) {
			return null;
		}
		return content.replace("&amp;" , "&" )
					  .replace("&lt;"  , "<" )
					  .replace("&gt;"  , ">" )
					  .replace("&quot;", "\"")
					  .replace("&apos;", "'" );
	}
	
	/**
	 * ɑ΂ JSON ̃GXP[vs
	 * @param content Ώە (null)
	 * @return GXP[v
	 * @since NicoCache_nl+120609mod
	 */
	public static String escapeJSON(String content) {
		if (content == null) {
			return null;
		}
		StringBuilder sb = new StringBuilder(content.length() * 6);
		char c;
		for (int i = 0; i < content.length(); i++) {
			c = content.charAt(i);
			if (c > 127) {
				sb.append(String.format("\\u%04x", (int) c));
			} else {
				switch (c) {
				case '/' : sb.append("\\/" ); break;
				case '"' : sb.append("\\\""); break;
				case '\\': sb.append("\\\\"); break;
				case '\b': sb.append("\\b" ); break;
				case '\f': sb.append("\\f" ); break;
				case '\n': sb.append("\\n" ); break;
				case '\r': sb.append("\\r" ); break;
				case '\t': sb.append("\\t" ); break;
				default  : sb.append(c);
				}
			}
		}
		return sb.toString();
	}
	
	/**
	 * JSON ̃GXP[vɖ߂
	 * @param content Ώە (null)
	 * @return GXP[v
	 * @since NicoCache_nl+120609mod
	 */
	public static String unescapeJSON(String content) {
		if (content == null) {
			return null;
		}
		StringBuffer sb = new StringBuffer(content.length());
		Matcher m = ASCII_UTF16_PATTERN.matcher(content);
		while (m.find()) {
			m.appendReplacement(sb, new String(
					Character.toChars(Integer.parseInt(m.group(1), 16))));
		}
		m.appendTail(sb);
		
		char c; String s;
		for (int i = 0; i < sb.length(); i++) {
			if (sb.charAt(i) == '\\' && i + 1 < sb.length()) {
				switch (c = sb.charAt(i + 1)) {
				case 'b': s = "\b"; break;
				case 'f': s = "\f"; break;
				case 'n': s = "\n"; break;
				case 'r': s = "\r"; break;
				case 't': s = "\t"; break;
				default : s = String.valueOf(c);
				}
				sb.replace(i, i + 2, s);
			}
		}
		return sb.toString();
	}
	
	
	private static final Pattern ASCII_UTF16_PATTERN = Pattern.compile(
			"(?<!\\\\)(?:\\\\\\\\)*\\\\u([0-9A-Fa-f]{4})");
	private static final Pattern ASCII_ESCAPE_PATTERN = Pattern.compile(
			"(?<!\\\\)(?:\\\\\\\\)*\\\\([/<>'\"])");
	
	/**
	 * \\uXXXX ɃGXP[vꂽ UTF-16 ɖ߂(native2ascii̋t)<br>
	 * X <b>/ < > ' "</b> ̃GXP[vɖ߂(JSONdlɏĂ܂)
	 * 
	 * @param content Ώە (null)
	 * @return fR[h
	 * @see #unescapeJSON(String)
	 */
	public static String ascii2native(String content) {
		if (content == null) {
			return null;
		}
		StringBuffer sb = new StringBuffer(content.length());
		Matcher m = ASCII_UTF16_PATTERN.matcher(content);
		while (m.find()) {
			m.appendReplacement(sb, new String(
					Character.toChars(Integer.parseInt(m.group(1), 16))));
		}
		m.appendTail(sb);
		return ASCII_ESCAPE_PATTERN.matcher(sb.toString()).replaceAll("$1");
	}
	
	/**
	 * Sp̉p𔼊pɕϊ
	 * 
	 * @param content ϊΏۂ̕ (null)
	 * @return ϊꂽ
	 */
	public static String zen2han(String content) {
		StringBuilder sb = new StringBuilder(content);
		char c;
		for (int i = 0; i < sb.length(); i++) {
			c = sb.charAt(i);
			if ('O' <= c && c <= 'X') {
				sb.setCharAt(i, (char) (c - 'O' + '0'));
			} else if ('`' <= c && c <= 'y') {
				sb.setCharAt(i, (char) (c - '`' + 'A'));
			} else if ('' <= c && c <= '') {
				sb.setCharAt(i, (char) (c - '' + 'a'));
			}
		}
		return sb.toString();
	}
	
	public static String join(String[] split) {
		return join(split, null, 0);
	}
	
	public static String join(String[] split, String separator) {
		return join(split, separator, 0);
	}
	
	public static String join(String[] split, String separator, int limit) {
		return join(Arrays.asList(split), separator, limit);
	}
	
	public static String join(List<String> split) {
		return join(split, null, 0);
	}
	
	public static String join(List<String> split, String separator) {
		return join(split, separator, 0);
	}
	
	/**
	 * ꂽA(JavaScript ̓\bhƓ)
	 * 
	 * @param split ꂽ܂ރXg
	 * @param separator ؂蕶Anullw肵ꍇ̓J}(",")
	 * @param limit A镶̏AftHg͑SĘA
	 * @return AꂽAXg̏ꍇ͋󕶎Ԃ
	 */
	public static String join(List<String> split, String separator, int limit) {
		if (separator == null) {
			separator = ",";
		}
		if (limit == 0 || split.size() < limit) {
			limit = split.size();
		}
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < limit; i++) {
			if (split.get(i) != null) {
				if (sb.length() > 0) {
					sb.append(separator);
				}
				sb.append(split.get(i));
			}
		}
		return sb.toString();
	}
	
	/**
	 * oCgPʂ̐lłK؂ȒPʂtɕϊ
	 * 
	 * @param bytes oCgPʂ̐l
	 * @return łK؂ȒPʂt
	 */
	public static String bytesToString(long bytes) {
		if (bytes > 1099511627776L) {
			return String.format("%,.3f TB", bytes / 1099511627776.0);
		} else if (bytes > 1073741824L) {
			return String.format("%,.2f GB", bytes / 1073741824.0);
		} else if (bytes > 1048576L) {
			return String.format("%,.1f MB", bytes / 1048576.0);
		} else if (bytes > 1024L) {
			return String.format("%,.1f KB", bytes / 1024.0);
		} else if (bytes > 1L) {
			return bytes + " bytes";
		}
		return bytes + " byte";
	}
	
}
