package dareka.processor.impl;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.util.concurrent.Executor;
import java.util.concurrent.FutureTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.Logger;
import dareka.extensions.SystemEventListener;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.SwfConvertResource;
import dareka.processor.impl.NicoIdInfoCache.Entry;

public class NicoCachingProcessor implements Processor {
    private static final String[] SUPPORTED_METHODS = new String[] { "GET" };
    /**
     * SMILEVIDEO̓URL̐K\
     */
    // http://s-clb5.smilevideo.jp/smile?v=102982.92382
    // 2009/3 http://smile-clb51.nicovideo.jp/smile?s=7074214.90075as3
    private static final Pattern SM_FLV_PATTERN =
//            Pattern.compile("^http://[^/]+(?:smilevideo|nicovideo)\\.jp/smile\\?(\\w)=([^.]+)\\.\\d+(as3)?(low)?$");
            Pattern.compile("^http://(?!tn-skr)[^/]+(?:smilevideo|nicovideo)\\.jp/smile\\?(\\w)=([^.]+)\\.\\d+(as3)?(low)?$");

    private Executor executor;

    public NicoCachingProcessor(Executor executor) {
        this.executor = executor;
    }

    public String[] getSupportedMethods() {
        return SUPPORTED_METHODS;
    }

    public Pattern getSupportedURLAsPattern() {
        return SM_FLV_PATTERN;
    }

    public String getSupportedURLAsString() {
        return null;
    }

    public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
            throws IOException {
        Matcher m = SM_FLV_PATTERN.matcher(requestHeader.getURI());
        if (!m.find()) {
            // it must not happen...
            return Resource.get(Resource.Type.URL, requestHeader.getURI());
        }

        MovieData data = new MovieData(m);
        String postfix = data.getCache().getPostfix();

        // [nl] mȂ`̓LbVȂ
        if (postfix.equals(".unknown")) {
        	Logger.warning("unknown data type: " + requestHeader.getURI());
        	return Resource.get(Resource.Type.URL, requestHeader.getURI());
        }

        // [nl] ExtensionɃLbVvCxgʒm
        NLEventSource eventSource = null;
        if (NLShared.INSTANCE.countSystemEventListeners() > 0) {
        	eventSource = new NLEventSource(null, requestHeader, data.getCache());
        	int result = NLShared.INSTANCE.notifySystemEvent(
        			SystemEventListener.CACHE_REQUEST, eventSource, true);
        	if (result != SystemEventListener.RESULT_OK) {
        		Logger.debug(requestHeader.getURI() + " pass-through by extension");
        		return Resource.get(Resource.Type.URL, requestHeader.getURI());
        	}
        }

// getflvRewriterŁAʏLbVƂ"low"폜Ă
// Aeco=1͍̎폜Ȃ̂ŁA
// ʏLbVlow̃ANZXeco=1Ɣf
        if (data.getCache().exists() && !(data.isLowAccess() && !data.getCache().isLow())) {
            Logger.info("using cache: " + data.getCache().getCacheFileName());
            if (Boolean.getBoolean("touchCache")) {
                data.getCache().touch();
            }
            Resource r;
            if (Boolean.getBoolean("swfConvert") && postfix.equals(Cache.SWF)) {
            	// [nl] LbVĐSWFϊs
            	SwfConvertResource swf = new SwfConvertResource(data.getCache().getURLString());
            	swf.setAS3(data.isAS3());
            	r = swf;
            } else {
            	r = Resource.get(Resource.Type.URL,
                            data.getCache().getURLString());
                // swfȊO̓uEUɃLbV
            	r.addCacheControlResponseHeaders(12960000);
            }
            r.setResponseHeader(HttpHeader.CONTENT_TYPE, "video/flv");
            LimitFlvSpeedListener.addTo(r);

            return r;
        }

// LbV̂ɂɗ̂eco=1̎
// łɒʏLbV̂ŃLbVȂ
// ꎞLbV͂AW[̓uEUɔCĂ̂܂ܕԂ
        if (data.getCache().exists()) {
        	Logger.info("force economy mode: " + data.getCache().getCacheFileName());
        	return Resource.get(Resource.Type.URL, requestHeader.getURI());
        }

// ĂяoꂽꍇɃLbVȂ
        // ĂяoŌÂXĝŁAKĂяo
        boolean isLive = NLShared.INSTANCE.isLiveMovie(data.getType() + data.getId());
        if (Boolean.getBoolean("noLiveCache")) {
        	// RefererŒe΁A̕m
        	String referer = requestHeader.getMessageHeader("Referer");
        	if (referer != null) {
        		isLive = referer.startsWith("http://live.nicovideo.jp") ||
        				 referer.contains("/liveplayer.swf");
        	}
        	if (isLive) {
        		return Resource.get(Resource.Type.URL, requestHeader.getURI());
        	}
        }

        // [nl] ExtensionLbV֎~ꍇ̓LbVȂ
        if (eventSource != null) {
        	int result = NLShared.INSTANCE.notifySystemEvent(
        			SystemEventListener.CACHE_STARTING, eventSource, true);
        	if (result != SystemEventListener.RESULT_OK) {
        		Logger.info("disable cache: " + data.getCache().getCacheFileName());
        		return Resource.get(Resource.Type.URL, requestHeader.getURI());
        	}
        }

        FutureTask<String> retrieveTitlteTask = null;
        if (Boolean.getBoolean("title")
                && (data.getIdInfo() == null || !data.getIdInfo().isTitleValid())) {
            retrieveTitlteTask =
                    new FutureTask<String>(new NicoCachingTitleRetriever(
                            data.getType(), data.getId()));
            executor.execute(retrieveTitlteTask);
        }

        Logger.info("no cache found: " + data.getCache().getCacheFileName());

        boolean dupconnect = Cache.getDLFlag(data.getCache().getId());
        long tmpSize = data.getCache().tmpLength();
        // dDLcacheAllocateFirstL
        // t@CTCY擾ς݃TCY
        if (dupconnect) {
        	tmpSize = Cache.getDLSize(data.getCacheId());
        }

        // [nl] uEŨW[ƁANCł̃W[Ή
        requestHeader.removeMessageHeader("Range");
        requestHeader.removeMessageHeader("If-Range");
        InputStream cacheInput = null;
        // [nl] dDLɂȂꍇ́A2߂̃RlNVł͎Ă镪̃LbV𗘗p
        // ܂A2߂̃RlNV͕ۑȂBAIɃW[
        // TODO:r邩ȂA܂͂
        // [nl] W[[
        if (tmpSize != 0
                && (dupconnect || Boolean.getBoolean("resumeDownload"))) {
            requestHeader.setMessageHeader("Range", "bytes=" + tmpSize + "-");
            cacheInput =
                    new BufferedInputStream(data.getCache().getTmpInputStream());
        }

        Resource r;
        try { // ensure cacheInput.close() in error cases.
            // [nl] DLXgɓ
        	if (!dupconnect) {
        		Cache.setDLFlag(data.getCache().getId(), 0);
        	}

            r = Resource.get(Resource.Type.URL, requestHeader.getURI());

            r.addTransferListener(new NicoCachingListener(data.getCache(),
                    retrieveTitlteTask, cacheInput, dupconnect, eventSource));
        } catch (RuntimeException e) {
            Logger.error(e);
            CloseUtil.close(cacheInput);
            Cache.setDLFlag(data.getCache().getId(), -1);
            throw e;
        }

        return r;
    }

    /**
     * Class for manage various data for a movie.
     *
     */
    static class MovieData {
        private String format;
        private String id;
        private String as3;
        private String suffix;
        private Entry idInfo;
        private String postfix;
        private String type;
        private Cache cache;

        MovieData(Matcher m) {
            initializeFormatIdSuffix(m);
            initializeIdInfo(getId());
            initializePostfix(format);
            initializeTypeCache(getIdInfo(), format);

            useAlternateCacheIfNecessary();
        }

        private void initializeFormatIdSuffix(Matcher m) {
            format = m.group(1);
            id = NLShared.INSTANCE.vid2cid(m.group(2)); // [nl]
            as3 = m.group(3);
            suffix = m.group(4);
        }

        private void initializeIdInfo(String id) {
            if (Boolean.getBoolean("title")) {
                idInfo = NicoIdInfoCache.getInstance().get(id);
            } else {
                idInfo = null;
            }
        }

        private void initializePostfix(String format) {
            if (format.equals("v")) {
                postfix = ".flv";
            } else if (format.equals("m")) {
                postfix = ".mp4";
            } else if (format.equals("s")) {
                postfix = ".swf";
            } else {
                postfix = ".unknown";
            }
        }

        private void initializeTypeCache(Entry idInfo, String format) {
            if (idInfo == null) {
                // this may not be correct, but we can not know without idInfo...
                if (format.endsWith("s")) {
                    type = "nm";
                } else {
                    type = "sm";
                }
                cache = new Cache(type + id, postfix);
            } else {
                type = idInfo.getType();
                cache = new Cache(type + id, postfix, idInfo.getTitle());
            }
        }

        private void useAlternateCacheIfNecessary() {
            if (!getCache().exists() && suffix != null) {
                if (idInfo == null) {
                    cache = new Cache(type + id + suffix, postfix);
                } else {
                    cache =
                            new Cache(type + id + suffix, postfix,
                                    idInfo.getTitle());
                }
            }
        }

        public Cache getCache() {
            return cache;
        }

        public Entry getIdInfo() {
            return idInfo;
        }

        public String getType() {
            return type;
        }

        public String getId() {
            return id;
        }

        public boolean isAS3() {
            return as3 != null;
        }

        public String getCacheId() {
            return (suffix == null) ? type + id : type + id + suffix;
        }

        public boolean isLowAccess() {
            return suffix != null;
        }
    }

}
